function [DATA,File,ChTbl,PSD,SPGRM]=PSD_SPGRM(DATA,File,ChTbl,PSD_WndwSec,PSD_OvlpPct,SPGRM_WndwSec,SPGRM_OvlpPct,AppResp,AppRespChs,DiffIntg,DiffIntgChs)
% Calculate SPEC(Spectrum), PSD(Power Spectrum Density) and Spectrogram, and Differentiate or Integrate DATA, SPEC, PSD, SPGRM
%
% Usage : [DATA,File,ChTbl,PSD,SPGRM] = PSD_SPGRM(DATA,File,ChTbl,PSD_WndwSec,PSD_OvlpPct,SPGRM_WndwSec,SPGRM_OvlpPct,DiffIntg);
%
% *** INPUT Parameters ***
% DATA,File,ChTbl : See ReadWIN.m
% PSD_WndwSec     : Window length in second for the PSD calculation
% PSD_OvlpPct     : Percentage of overlaps of windowed sections for the PSD calculation
% SPGRM_WndwSec   : Window length in second for the spectrogram calculation
% SPGRM_OvlpPct   : Percentage of overlaps of windowed sections for the spectrogram calculation
% AppResp         : Apply sensor response based on the ChTbl.Damp and ChTbl.NatFrq. 'Yes' or 'No' or []. Default is 'No'.
% AppRespChs      : Target channel ID(s) that ApplyResp is applied (cell array). Ex) {'00A0','00B1'} or {} (All the channels are applied)
% DiffIntg        : 'Differencial' or 'Integral' or 'none' or '' or [] (Empty is the same as 'none')
% DiffIntgChs     : Target channel ID(s) that DiffIntg is applied (cell array). Ex) {'00A0','00B1'} or {} (All the channels are applied)
%
% *** OUTPUT Parameters ***
% DATA,File,ChTbl : See ReadWIN.m
% PSD             : PSD (struct array). Column of PSD.PSD corresponds to channel ID(s).
% SPGRM           : Spectrogram data (struct array)
%
% *** UPDATES ***
% 2008/July/18(Fri) : Original by Koji MIYAKAWA
%

% ---------------------------
% Parameter Initialization
%
if exist('PSD_WndwSec','var') && ~isempty(PSD_WndwSec)      % Check PSD_WndwSec
    File.PSD_WndwSec = PSD_WndwSec;
else
    File.PSD_WndwSec = [];                                  % Use default window length (data is divided into eight time windows)
end
if exist('PSD_OvlpPct','var') && ~isempty(PSD_OvlpPct)      % Check PSD_OvlpPct
    File.PSD_OvlpPct = PSD_OvlpPct;
else
    File.PSD_OvlpPct = [];                                  % Use default overlap (50%)
end
if exist('SPGRM_WndwSec','var') && ~isempty(SPGRM_WndwSec)  % Check SPGRM_WndwSec
    File.SPGRM_WndwSec = SPGRM_WndwSec;
else
    File.SPGRM_WndwSec = [];                                % Use default window length (data is divided into eight time windows)
end
if exist('SPGRM_OvlpPct','var') && ~isempty(SPGRM_OvlpPct)  % Check SPGRM_OvlpPct
    File.SPGRM_OvlpPct = SPGRM_OvlpPct;
else
    File.SPGRM_OvlpPct = [];                                % Use default overlap (50%)
end
if exist('AppResp','var') && strcmpi(AppResp,'Yes')
    if exist('AppRespChs','var') && ~isempty(AppRespChs)    % Check AppRespChs
        File.AppRespChs = AppRespChs;
    else
        File.AppRespChs = ChTbl.ChID;
    end
end
if exist('DiffIntg','var') && (strcmpi(DiffIntg,'Differential') || strcmpi(DiffIntg,'Integral'))
    if exist('DiffIntgChs','var') && ~isempty(DiffIntgChs)    % Check DiffIntgChs
        File.DiffIntgChs = DiffIntgChs;
    else
        File.DiffIntgChs = ChTbl.ChID;
    end
end

% ---------------------------
% ApplySensorResponse and Differentiation/integration
%
if (exist('DiffIntg','var') && (strcmpi(DiffIntg,'Differential') || strcmpi(DiffIntg,'Integral'))) || (exist('AppResp','var') && strcmpi(AppResp,'Yes'))
    for ii = 1:length(DATA)
        SPEC     = fft(DATA{ii})/size(DATA{ii},1);          % Caluculate ci = (ai-j*bi)/2. f = (0:size(SPEC,1)-1)'/File.RecSec;
        EndPoint = fix(size(SPEC,1)/2)+rem(size(SPEC,1),2); % End Point Number for One Sided Amplitude/PSD
        % PSD.F   = (0:size(SPEC,1)-1)'/File.RecSec;        % Frequency
        % PSD.PSD = abs(SPEC).^2*File.RecSec;               % Standard Two Sided PSD Estimation
        % PSD.PSD = PSD(1:fix(size(PSD,1)/2)+1,:);          % Chop Conjugate Part
        % PSD.PSD(2:EndPoint,:) = PSD(2:EndPoint,:)*2;      % One Sided PSD estimated from Spectrum Data
        SPEC     = SPEC(1:fix(size(SPEC,1)/2)+1,:);         % Chop Conjugate Part
        SPEC(2:EndPoint,:) = SPEC(2:EndPoint,:)*2;          % Double Amplitude to make it One Sided Amplitude = One Sided Spectrum
        if exist('AppResp','var') && strcmpi(AppResp,'Yes') % Apply sensor response
            F       = (0:size(SPEC,1)-1)'/File.RecSec;
            FlatPts = find(F<0.01,1,'last');                % Sensor response is applied up to 0.01Hz
            for jj = 1:size(SPEC,2)                         % Each channel
                TarCh = File.StChs(ii)+jj-1;
                if any(hex2dec(File.AppRespChs) == hex2dec(ChTbl.ChID{TarCh}))
                    RATIO = -F.^2./(F.^2-i*(2*ChTbl.Damp(TarCh)*ChTbl.NatFrq(TarCh))*F-ChTbl.NatFrq(TarCh)^2);
                    if isempty(FlatPts)
                        RATIO(1) = 1;
                    else
                        RATIO(1:FlatPts) = 1;               % Make flat response lower than 0.1Hz
                    end
                    SPEC(:,jj) = SPEC(:,jj)./RATIO;
                end
            end
        end
        if exist('DiffIntg','var') && strcmpi(DiffIntg,'Differential')  % Differentiation
            for jj = 1:size(SPEC,2)                         % Each channel
                TarCh = File.StChs(ii)+jj-1;
                if any(hex2dec(File.DiffIntgChs) == hex2dec(ChTbl.ChID{TarCh}))
                    SPEC(:,jj) = SPEC(:,jj).*(i*2*pi*(0:size(SPEC,1)-1)'/File.RecSec);
                    if strcmpi(ChTbl.Unit{TarCh},'m/s/s')
                        ChTbl.Unit{TarCh} = 'm/s/s/s';
                    elseif strcmpi(ChTbl.Unit{TarCh},'m/s')
                        ChTbl.Unit{TarCh} = 'm/s/s';
                    elseif strcmpi(ChTbl.Unit{TarCh},'m')
                        ChTbl.Unit{TarCh} = 'm/s';
                    else
                        ChTbl.Unit{TarCh} = [ChTbl.Unit{TarCh},'/s'];
                    end
                end
            end
            File.DiffIntg = 'Differentiated';
        elseif exist('DiffIntg','var') && strcmpi(DiffIntg,'Integral')  % Integration
            for jj = 1:size(SPEC,2)                         % Each channel
                TarCh = File.StChs(ii)+jj-1;
                if any(hex2dec(File.DiffIntgChs) == hex2dec(ChTbl.ChID{TarCh}))
                    SPEC(2:end,jj) = SPEC(2:end,jj)./(i*2*pi*(1:size(SPEC,1)-1)'/File.RecSec);
                    if strcmpi(ChTbl.Unit{TarCh},'m/s/s')
                        ChTbl.Unit{TarCh} = 'm/s';
                    elseif strcmpi(ChTbl.Unit{TarCh},'m/s')
                        ChTbl.Unit{TarCh} = 'm';
                    elseif strcmpi(ChTbl.Unit{TarCh},'m')
                        ChTbl.Unit{TarCh} = 'm*s';
                    else
                        ChTbl.Unit{TarCh} = [ChTbl.Unit{TarCh},'*s'];
                    end
                end
            end
            File.DiffIntg = 'Integrated';
        end
        DATA{ii} = real(ifft(SPEC,size(DATA{ii},1)))*size(DATA{ii},1);
    end
    if exist('AppResp','var') && strcmpi(AppResp,'Yes') % Apply sensor response
        fprintf('Sensor response is applied            :');
        for ii = 1:length(File.AppRespChs)
            fprintf(' %s(%GHz)',File.AppRespChs{ii},File.SmpFrq(File.ColChs(hex2dec(ChTbl.ChID)==hex2dec(File.AppRespChs{ii}))));
        end
        fprintf('\n');
    end
    if exist('DiffIntg','var') && strcmpi(DiffIntg,'Differential')  % Differentiation
        fprintf('DATA/SPEC/PSD/SPGRM are differentiated:');
        for ii = 1:length(File.DiffIntgChs)
            fprintf(' %s(%GHz)',File.DiffIntgChs{ii},File.SmpFrq(File.ColChs(hex2dec(ChTbl.ChID)==hex2dec(File.DiffIntgChs{ii}))));
        end
        fprintf('\n');
    elseif exist('DiffIntg','var') && strcmpi(DiffIntg,'Integral')  % Integration
        fprintf('DATA/SPEC/PSD/SPGRM are integrated    :');
        for ii = 1:length(File.DiffIntgChs)
            fprintf(' %s(%GHz)',File.DiffIntgChs{ii},File.SmpFrq(File.ColChs(hex2dec(ChTbl.ChID)==hex2dec(File.DiffIntgChs{ii}))));
        end
        fprintf('\n');
    end
    fprintf('\n');
end

% ---------------------------
% One Sided PSD estimated using pwelch function
%
PSD = cell(length(DATA),1);
for ii = 1:length(DATA)
    if isempty(File.PSD_WndwSec) || File.PSD_WndwSec*File.SmpFrq(ii) < 60         % WINDOW needs more than 60 data points.
        WINDOW = [];
    else
        WINDOW = File.PSD_WndwSec*ceil(File.SmpFrq(ii));
    end
    for jj = 1:size(DATA{ii},2)
        [PSD{ii}.PSD(:,jj),PSD{ii}.F] = pwelch(DATA{ii}(:,jj)-mean(DATA{ii}(:,jj)),WINDOW,File.PSD_OvlpPct/100*WINDOW,[],File.SmpFrq(ii));                                   % DC Removal only
%       [PSD{ii}.PSD(:,jj),PSD{ii}.F] = pwelch(DATA{ii}(:,jj)-polyval(polyfit((0:size(DATA{ii},1)-1)'/File.SmpFrq(ii),DATA{ii}(:,jj),1),(0:size(DATA{ii},1)-1)'/File.SmpFrq(ii)),WINDOW,File.PSD_OvlpPct/100*WINDOW,[],File.SmpFrq(ii)); % DC & Trend Removal
    end
end

% ---------------------------
% One Sided PSD Spectrogram
%
SPGRM = cell(length(DATA),1);
for ii = 1:length(DATA)
    if isempty(File.SPGRM_WndwSec) || File.SPGRM_WndwSec*File.SmpFrq(ii) < 60         % WINDOW needs more than 60 data points.
        WINDOW = [];
    else
        WINDOW = File.SPGRM_WndwSec*ceil(File.SmpFrq(ii));
    end
    for jj = 1:size(DATA{ii},2)
        [S,SPGRM{ii}(jj).F,SPGRM{ii}(jj).T,SPGRM{ii}(jj).PSD] = spectrogram(DATA{ii}(:,jj)-mean(DATA{ii}(:,jj)),WINDOW,File.SPGRM_OvlpPct/100*WINDOW,[],File.SmpFrq(ii));                                   % DC Removal only
%       [S,SPGRM{ii}(jj).F,SPGRM{ii}(jj).T,SPGRM{ii}(jj).PSD] = spectrogram(DATA{ii}(:,jj)-polyval(polyfit((0:size(DATA{ii},1)-1)'/File.SmpFrq(ii),DATA{ii}(:,ii),1),(0:size(DATA{ii},1)-1)'/File.SmpFrq(ii)),WINDOW,SPGRM_OvlpPct/100*WINDOW,[],File.SmpFrq(ii)); % DC & Trend Removal
    end
end

% ---------------------------
% Another differential/integral method based on time series
% Only DATA parameter is differentiated/integrated.
%
% if exist('DiffIntg','var') && strcmpi(DiffIntg,'Differential')
%     fprintf('DATA is differentiated.\n\n');
%     for ii = 1:length(DATA)
%         for jj = 1:size(DATA{ii},2)
%             TarCh = File.StChs(ii)+jj-1;
%             DATA{ii}(:,jj)    = gradient(DATA{ii}(:,jj),1/File.SmpFrq(ii));
%             ChTbl.Unit{TarCh} = ['Differential of ',ChTbl.Unit{TarCh}];
%         end
%     end
% elseif exist('DiffIntg','var') && strcmpi(DiffIntg,'Integral')
%     fprintf('DATA is Integrated.\n\n');
%     for ii = 1:length(DATA)
%         for jj = 1:size(DATA{ii},2)
%             TarCh = File.StChs(ii)+jj-1;
%             DATA{ii}(:,jj)    = cumtrapz((0:size(DATA{ii},1)-1)/File.SmpFrq(ii),DATA{ii}(:,jj));
%             DATA{ii}(:,jj)    = DATA{ii}(:,jj)-polyval(polyfit((0:size(DATA{ii},1)-1)/File.SmpFrq(ii),DATA{ii}(:,jj),1),(0:size(DATA{ii},1)-1)/File.SmpFrq(ii)); % Trend Removal
%             ChTbl.Unit{TarCh} = ['Integral of ',ChTbl.Unit{TarCh}];
%         end
%     end
% end
%
