function Plot_WVs(DATA,File,ChTbl,Xunit,XLIM,YRange,PlotSkip)
% Plot multiple waveforms
%
% Usage : Plot_WVs(DATA,File,ChTbl,Xunit,XLIM,YLIM,PlotSkip)
%
% *** INPUT Parameters ***
% DATA,File,ChTbl : See ReadWIN.m
% Xunit      : Unit of Xaxis. 'Sec' or 'Minute' or 'Hour'. Default is 'Sec'. Default is 'Sec'.
% XLIM       : Xaxis limits (time). [Xmin Xmax] or [] (Display whole range). If Xunit is 'Clock', DataXLIM must be date datevec format like {[yyyy mm dd HH MM SS],[yyyy mm dd HH MM SS]}.
% YRange     : Yaxis range (One-sided amplitude from the YMEAN). Default is whole range.
% PlotSkip   : Number of skipping data points for waveform plot. Default is 1 (no skip).
%
% *** UPDATES ***
% 2008/July/18(Fri) : Original by Koji MIYAKAWA
%

clf;
% ---------------------------
% Parameter Initialization
%
RECT  = [0.08 0.05 0.88 0.94];      % RECT = get(gcf,'DefaultAxesPosition');
NumCh = sum(File.NumChs);
ha    = zeros(NumCh,1);
if exist('Xunit','var') && strcmpi(Xunit,'Minute')      % Check Xunit
    for ii = 1:length(File.NumChs)
        TIME{ii} = (0:size(DATA{ii},1)-1)'/File.SmpFrq(ii)/60;    % Minute
    end
    Xlabel    = '/Min';
elseif exist('Xunit','var') && strcmpi(Xunit,'Hour')    % Hour
    for ii = 1:length(File.NumChs)
        TIME{ii} = (0:size(DATA{ii},1)-1)'/File.SmpFrq(ii)/3600;
    end
    Xlabel    = '/Hour';
elseif exist('Xunit','var') && strcmpi(Xunit,'Clock')   % Clock
    for ii = 1:length(File.NumChs)
        TIME{ii} = (0:size(DATA{ii},1)-1)'/File.SmpFrq(ii)/86400+File.StNum;
    end
    if exist('XLIM','var') && ~isempty(XLIM)
        XLIM  = [datenum(XLIM{1}) datenum(XLIM{2})];
    end
    Xlabel    = '';
else                                                    % Sec
    for ii = 1:length(File.NumChs)
        TIME{ii} = (0:size(DATA{ii},1)-1)'/File.SmpFrq(ii);
    end
    Xunit     = 'Sec';
    Xlabel    = '/Sec';
end
if ~exist('XLIM','var') || isempty(XLIM)                % Check XLIM
    XLIM = [TIME{1}(1) TIME{1}(end)+TIME{1}(2)-TIME{1}(1)];
end
if ~exist('YRange','var') || isempty(YRange)            % Check YRange
    for ii = 1:length(File.NumChs)
        YRange(ii) = max(max(DATA{ii})-min(DATA{ii}));
    end
%   YRange = ones(1,length(File.NumChs))*max(YRange);   % Set the same YLIM in all axeses
else
    YRange = ones(1,length(File.NumChs))*YRange;
end
if exist('File','var') && isfield(File,'BPFfc') && ~isempty(File.BPFfc) % Check File.BPFfc
    if isnan(File.BPFfc(1))
        FiltStr = ['LPF=',num2str(File.BPFfc(2)),'Hz'];
    elseif isnan(File.BPFfc(2))
        FiltStr = ['HPF=',num2str(File.BPFfc(1)),'Hz'];
    else
        FiltStr = ['BPF=',num2str(File.BPFfc(1)),'-',num2str(File.BPFfc(2)),'Hz'];
    end
else
    FiltStr = 'No Filter';
end
if ~exist('PlotSkip','var') || isempty(PlotSkip)        % Check PlotSkip
    PlotSkip = 1;
end

% ---------------------------
% Plot
%
for ii = 1:NumCh
    TarCol    = File.ColChs(ii);
    TarRow    = File.RowChs(ii);
    YMEAN(ii) = mean(DATA{TarCol}(1:PlotSkip:end,TarRow));
    ha(ii)    = axes('Position',[RECT(1,1)+0.015 RECT(1,2)+RECT(1,4)*(NumCh-ii)/NumCh RECT(1,3) RECT(1,4)/NumCh*0.93]);
    plot(TIME{TarCol}(1:PlotSkip:end),DATA{TarCol}(1:PlotSkip:end,TarRow)-YMEAN(ii)); grid on;
    ylabel({sprintf('Amplitude/%s',ChTbl.Unit{ii});sprintf('DCoffset=\\bf%G\\rm)',YMEAN(ii))});
    legend(ha(ii),sprintf('\\bf%s-%s (%s)\\rm',strrep(ChTbl.StnNm{ii},'_','\_'),ChTbl.Comp{ii},ChTbl.ChID{ii}));
    xlim(XLIM);
    if strcmpi(Xunit,'Clock')
        datetick(ha(ii),'x','HH:MM','keeplimits');
    else
        DoubleTick(ha(ii),'x');     % Double the number of the XTick
    end
    ylim([-YRange(TarCol)/2 YRange(TarCol)/2]);
    if ii == NumCh
        xlabel(sprintf('Time%s (%s, PlotSkip=%dpts)',Xlabel,FiltStr,PlotSkip));
    end
end
set(ha,'Box','on');

% ---------------------------
% Text Information
%
CurAxes = gca;
axes('Position',[0 0 1 1],'Visible','off');
if File.RecSec < 3*3600                         % Set 'minute' if the data length is less than 3 hours
    DUR = [num2str(round(File.RecSec/60*10)/10),'min'];
else
    DUR = [num2str(round(File.RecSec/3600*10)/10),'hr'];
end
TITLE = sprintf('%s %s - %s (%s)',File.StDate,File.StTime,File.EdTime,DUR);
text(0.5,0.998,TITLE,'HorizontalAlignment','center','VerticalAlignment','top','FontSize',12,'FontWeight','bold');
set(gcf,'CurrentAxes',CurAxes);

% ===========================
function DoubleTick(ha,Axis)
% Double the number of the Tick
%
if strcmpi(Axis,'x')
    LIMstr  = 'XLim';
    TICKstr = 'XTick';
elseif strcmpi(Axis,'y')
    LIMstr  = 'YLim';
    TICKstr = 'YTick';
else
    return
end
LIM   = get(ha,LIMstr);
TICK  = get(ha,TICKstr);
dTICK = (TICK(2)-TICK(1))/2;
if LIM(2) > TICK(end)+dTICK
    TICK(end) = TICK(end)+dTICK;
end
set(ha,TICKstr',TICK(1):dTICK:TICK(end));
