function [DATA,File,ChTbl] = ReadDAT(DATfn,ChTBLfn,ChIDcell,StEdTime)
% Read DAT5 Compression Data
%
% ---------------------------
% Initial Parameters
%
FORMAT   = {'ubit24','','ubit20','','ubit16','','ubit12','ubit8'};
SUBST    = [ 2^23   , 0, 2^19   , 0, 2^15   , 0, 2^11   , 2^7]; 
File.GPS = [];
DATA{1}  = [];

% ---------------------------
% Open File
%
File.name = DATfn;
[fid, message] = fopen(File.name,'r','b');
if fid == -1, error(message); end

% ---------------------------
% Read First Block Header
%
fread(fid,8,'ubit4');                           % (  4Bytes) 0x242424
File.SN          = fread(fid,  1,'ubit16');     % (  2Bytes) Serial Number
tmp              = (reshape(fread(fid, 16,'ubit4'),2,8)'*[10;1])';  % (  8Bytes) BCD
File.StNum       = datenum([tmp(1)+CheckYY00(tmp(1)),tmp(2:5),tmp(6)+tmp(7:8)*[0.01;0.0001]]);
File.SmpFrq      = round(10^6/fread(fid,  1,'ubit32')); % (  4Bytes) Sampling Rate (micro sec)
NumChs           = fread(fid,  1,'ubit16');  	% (  2Bytes) Number of Channels
fread(fid,  1,'ubit32');                        % (  4Bytes) Number of Sampling Data
File.model       = fread(fid,  1,'ubit16');     % (  2Bytes) Model Number
File.GPS         = ReadGPS(fid,File.GPS);       % (358Bytes) GPS status
if ~isempty(File.GPS) && any(File.GPS.valid == 1)
    ii = find(File.GPS.valid == 1,1);
    if any(ii)
        File.StNum = File.StNum+File.GPS.GPStime(ii)-File.GPS.RECtime(ii);
    end
end
File.GPS = [];
frewind(fid);

% ---------------------------
% ChID Setup
%
ChIDdecTmp = 1:NumChs;
if exist('ChIDcell','var') && ~isempty(ChIDcell)
    [~,Idx] = sort(hex2dec(ChIDcell));
    ChIDcell = ChIDcell(Idx);
    for ii = 1:length(ChIDcell)
        jj = find(ChIDdecTmp == hex2dec(ChIDcell{ii}),1);
        if isempty(jj)
            error(['Your specified ChID (',ChIDcell{ii},') does not exist in the DAT file.']);
        else
            ChIDdec(ii) = ChIDdecTmp(jj);
        end
    end
else
    ChIDdec = ChIDdecTmp;
end
File.NumChs = length(ChIDdec);

% ---------------------------
% Start/End Time Setup
%
fprintf('Start time of this DAT file : %04d/%02d/%02d %02d:%02d:%06.3f\n',datevec(File.StNum));
if exist('StEdTime','var') && ~isempty(StEdTime)        % Check StEdTime
    SubSecOffset = File.StNum*86400-fix(File.StNum*86400);
    if length(StEdTime) == 1                            % StEdTime = DataLength (sec)
        StPts = 1;
        EdPts = StEdTime*File.SmpFrq;
    elseif length(StEdTime{2}) == 1                     % StEdTime = {[StTime],DataLength(sec)}
        StNum = datenum(StEdTime{1})+SubSecOffset/86400;
        StPts = round((StNum-File.StNum)*86400*File.SmpFrq)+1;
        if StPts < 1
            StPts = 1;
        end
        EdPts = StPts+StEdTime{2}*File.SmpFrq-1;
        File.StNum  = StNum;
    else                                                % StEdTime = {[StTime],[EdTime]}
        StNum = datenum(StEdTime{1})+SubSecOffset/86400;
        EdNum = datenum(StEdTime{2})+SubSecOffset/86400;
        StPts = round((StNum-File.StNum)*86400*File.SmpFrq)+1;
        if StPts < 1
            StPts = 1;
        end
        EdPts = StPts+round((EdNum-StNum)*86400*File.SmpFrq)-1;
        File.StNum  = StNum;
    end
else
    StPts = 1;
    EdPts = Inf;
end
fprintf('Start time to be extracted  : %04d/%02d/%02d %02d:%02d:%06.3f\n',datevec(File.StNum));

% ---------------------------
% Read Channel Table File
%
if exist('ChTBLfn','var') && ~isempty(ChTBLfn)
    File.ChTbl = ChTBLfn;
    [fid1, message] = fopen(File.ChTbl,'r','n');
    if fid1 == -1, error(message); end
    while 1
        l = fgetl(fid1);
        if ~ischar(l), break, end
        L  = strread(l,'%s',18,'commentstyle','shell');
        if isempty(L), continue, end
        ii = find(ChIDdec == hex2dec(L{1}),1);
        if isempty(ii), continue, end
        if length(L) < 18
            L(length(L)+1:18,1) = cell(18-length(L),1);     % Zero Padding
        end
        ChTbl.ChID{ii}   = sprintf('%04X',hex2dec(L{1}));
        ChTbl.StnNm{ii}  = L{4};
        ChTbl.Comp{ii}   = L{5};
        ChTbl.ADbit(ii)  = str2double(L{7});
        ChTbl.Sensi(ii)  = str2double(L{8});
        ChTbl.Unit{ii}   = L{9};
        ChTbl.NatFrq(ii) = str2double(L{10});
        ChTbl.Damp(ii)   = str2double(L{11});
        ChTbl.dB(ii)     = str2double(L{12});
        ChTbl.ADres(ii)  = str2double(L{13});
        ChTbl.Lat(ii)    = str2double(L{14});
        ChTbl.Lon(ii)    = str2double(L{15});
        ChTbl.Alt(ii)    = str2double(L{16});
        ChTbl.Pcorr(ii)  = str2double(L{17});
        ChTbl.Scorr(ii)  = str2double(L{18});
    end
    fclose(fid1);
    if ~exist('ChTbl','var') || any(cellfun('isempty',ChTbl.ChID)) || ~all(hex2dec(ChTbl.ChID)'==ChIDdec)
        error('Specified ChID(s) are not found in the ChTbl file.');
    end
else
    File.ChTbl = '';
    for ii = 1:length(ChIDdec)
        ChTbl.ChID{ii}   = sprintf('%04X',ChIDdec(ii));
        ChTbl.StnNm{ii}  = num2str(File.SN);
        ChTbl.Comp{ii}   = num2str(ii);
        ChTbl.ADbit(ii)  = 24;
        ChTbl.Sensi(ii)  = 1;
        ChTbl.Unit{ii}   = 'V';
        ChTbl.NatFrq(ii) = 0;
        ChTbl.Damp(ii)   = 0;
        ChTbl.dB(ii)     = 0;
        ChTbl.ADres(ii)  = 4/2^23;
        ChTbl.Lat(ii)    = 0;
        ChTbl.Lon(ii)    = 0;
        ChTbl.Alt(ii)    = 0;
        ChTbl.Pcorr(ii)  = 0;
        ChTbl.Scorr(ii)  = 0;
    end
end

% ---------------------------
% Read Data
%
blk    = 1;
CurPts = 1;
while 1
    if CurPts > EdPts
        break;
    end
    while 1
        if feof(fid)
            break;
        elseif fread(fid,1,'ubit4') == 2     	% (  4Bytes) 0x24242424
            if all(fread(fid,7,'ubit4') == [4;2;4;2;4;2;4])
                break;
            end
        end
    end
    if feof(fid), break, end
% ---------------------------
% Read Block Header
%
    fread(fid,1,'ubit16');                            	% (  2Bytes) Serial Number
    tmp = fread(fid,16,'ubit4')';                       % (  8Bytes) BCD
    tmp = (reshape(tmp,2,8)'*[10;1])';
    File.StTm(blk) = datenum([tmp(1)+CheckYY00(tmp(1)),tmp(2:5),tmp(6)+tmp(7:8)*[0.01;0.0001]]);
    fprintf('BlockNum : %3d , HeaderTime : %04d/%02d/%02d %02d:%02d:%06.3f\n',blk,datevec(File.StTm(blk)));
    fread(fid,1,'ubit32');                              % (  4Bytes) Sampling Rate (micro sec)
    fread(fid,1,'ubit16');                              % (  2Bytes) Number of Channels
    File.NumSmp(blk) = fread(fid,1,'ubit32');           % (  4Bytes) Number of Sampling Data
    fread(fid,1,'ubit16');                              % (  2Bytes) Model Number
    File.GPS = ReadGPS(fid,File.GPS);                   % (358Bytes) GPS status

% ---------------------------
% Parameter Initialize
%
    if CurPts+File.NumSmp(blk) <= StPts
        BlkSt = File.NumSmp(blk);
        BlkEd = 0;
    else
        if CurPts >= StPts
            BlkSt = 1;
        else
            BlkSt = StPts-CurPts+1;
        end
        if CurPts+File.NumSmp(blk) <= EdPts
            BlkEd = File.NumSmp(blk);
        else
            BlkEd = EdPts-CurPts+1;
        end
        OFFSET = size(DATA{1},1);
        DATA{1}(OFFSET+1:OFFSET+BlkEd-BlkSt+1,1:File.NumChs) = 0;
    end
% ---------------------------
% Read Block Data
%
    BlkCur = 1;
%     fprintf('StPts=%d, EdPts=%d, CurPts=%d, BlkSt=%d, BlkEd=%d, BlkCur=%d\n',StPts,EdPts,CurPts,BlkSt,BlkEd,BlkCur);
    while 1
        if BlkCur > File.NumSmp(blk) || CurPts > EdPts
            break;
        end
        if BlkCur >= BlkSt && BlkCur <= BlkEd
            REC = 1;
        else
            REC = 0;
        end
        for ii = 1:NumChs
            ChHEAD = fread(fid,8,'ubit1')';
%             disp([num2str(BlkCur),' : ',num2str(ChHEAD)]);
            k = [ChHEAD(1:2),ChHEAD(4)]*[4;2;1]+1;  % 1:24bit, 3:20bit, 5:16bit, 7:12bit, 8:8bit
            Idx = find(ChIDdec == ChHEAD(5:8)*[8;4;2;1]+1,1);
            if REC && any(Idx)
                if ChHEAD(3) == 0
                    DATA{1}(OFFSET+BlkCur-BlkSt+1:OFFSET+BlkCur-BlkSt+File.SmpFrq,Idx) = cumsum([fread(fid,1,'ubit24')-2^23;(fread(fid,File.SmpFrq-1,FORMAT{k})-SUBST(k))*-1]); % Cumulative Sum
                elseif ChHEAD(3) == 1
                    DATA{1}(OFFSET+BlkCur-BlkSt+1:OFFSET+BlkCur-BlkSt+2,Idx) = fread(fid,2,'ubit24')-2^23;
                    for jj = OFFSET+BlkCur-BlkSt+3:OFFSET+BlkCur-BlkSt+File.SmpFrq
                        DATA{1}(jj,Idx) = (fread(fid,1,FORMAT{k})-SUBST(k))-DATA{1}(jj-2,Idx)+DATA{1}(jj-1,Idx)*2;
                    end
                else
                    error(['Compression Type is wrong : ',num2str(ChHEAD(3))]);
                end
            else
                if ChHEAD(3) == 0
                    fread(fid,1,'ubit24');
                    fread(fid,File.SmpFrq-1,FORMAT{k});
                elseif ChHEAD(3) == 1
                    fread(fid,2,'ubit24');
                    fread(fid,File.SmpFrq-2,FORMAT{k});
                else
                    error(['Compression Type is wrong : ',num2str(ChHEAD(3))]);
                end
            end
            if k == 3 && ChHEAD(3) == 0   	% 20bit
                fread(fid,1,'ubit20');
            elseif k == 7 && ChHEAD(3) == 0	% 12bit
             	fread(fid,1,'ubit12');
            end
        end
        BlkCur = BlkCur+File.SmpFrq;
        CurPts = CurPts+File.SmpFrq;
    end
    blk = blk+1;
end
fclose(fid);

% ---------------------------
% Make File Struct
%
[File.StVec,File.StDate,File.StTime] = Num2VecDateTime(File.StNum);
File.RecSec    = round(size(DATA{1},1)/File.SmpFrq);
File.EdNum     = File.StNum+File.RecSec/86400;
[File.EdVec,File.EdDate,File.EdTime] = Num2VecDateTime(File.EdNum);
File.StChs     = 1;
File.ColChs    = ones(1,File.NumChs);
File.RowChs    = 1:File.NumChs;
File.NoDATA{1} = [];

% ---------------------------
% Data Unit Conversion
%
for ii = 1:File.NumChs
	DATA{File.ColChs(ii)}(:,File.RowChs(ii)) = DATA{File.ColChs(ii)}(:,File.RowChs(ii))/10^(ChTbl.dB(ii)/20)*ChTbl.ADres(ii)/ChTbl.Sensi(ii);
end

% ---------------------------
% Clear unuseful variables
%
File = rmfield(File,{'SN','StTm','NumSmp','model'});
fprintf('\n');

% ---------------------------
% Plot
%
% figure(1); clf;
% TIME = (0:size(DATA{1},1)-1)'/File.SmpFrq/86400+File.StNum;
% for ii = 1:File.NumChs
%     subplot(File.NumChs,1,ii), plot(TIME,DATA{File.ColChs(ii)}(:,File.RowChs(ii))); grid on; legend(['Ch',ChTbl.Comp{ii}])
%     datetick('x','HH:MM','keeplimits'); xlabel('Time'); ylabel(ChTbl.Unit{ii});
%     if ii == 1
%         title([File.StDate,' ',File.StTime]);
%     end
% end
% keyboard

function GPS = ReadGPS(fid,GPS)
% Read GPS status from Header
%
tmp = fread(fid,358,'*char')';      % (358Bytes) GPS status
tmp(tmp==char(0))= ' ';
tmp = textscan(tmp,'%s');
tmp = tmp{1};
Idx = find(strcmp(tmp,'GPS'));
if ~isempty(Idx)
	if ~isempty(GPS)
        k = length(GPS.valid);
    else
        k = 0;
    end    
	for ii = 1:length(Idx)
        if strcmp(tmp{Idx(ii)+1},'VALID')
            GPS.valid(k+ii)   = 1;
            GPS.lat{k+ii}     = tmp{Idx(ii)+2};
            GPS.lon{k+ii}     = tmp{Idx(ii)+3};
            tmp2              = sscanf(tmp{Idx(ii)+4},'%2d')';
            GPS.GPStime(k+ii) = datenum([tmp2(1)+CheckYY00(tmp2(1)),tmp2(2:6)]);
            tmp2              = sscanf(tmp{Idx(ii)+5},'%2d')';
            GPS.RECtime(k+ii) = datenum([tmp2(1)+CheckYY00(tmp2(1)),tmp2(2:5),tmp2(6)+str2double(tmp{Idx(ii)+6})/1000000]);
        else
            GPS.valid(k+ii)   = 0;
            GPS.lat{k+ii}     = '';
            GPS.lon{k+ii}     = '';
            GPS.GPStime(k+ii) = 0;
            GPS.RECtime(k+ii) = 0;
        end
    end
end

% ===========================
function YY00 = CheckYY00(yy)
% Check YY00 is 1900 or 2000
%
if yy < 50
	YY00 = 2000;
else
    YY00 = 1900;
end

% ===========================
function [Vec,Date,Time]=Num2VecDateTime(Num)
% Convert a serial date number to vector, date string, and time string
%
Vec  = datevec(Num);
Date = [datestr(Num,'yyyy/mm/dd('),datestr(Num,'ddd)')];
Time = sprintf('%02d:%02d:%06.3f',Vec(4:6));
