function [DATA,File,ChTbl] = ReadWIN(WINfn,ChTBLfn,ChIDcell,StEdTime)
% Import WIN file data into Matlab variables
%
% Usage   : [DATA,File,ChTbl] = ReadWIN(WINfn,ChTBLfn,ChIDcell,StEdTime);
% Example : [DATA,File,ChTbl] = ReadWIN('C:\test.win','C:\channels.tbl');
% Example : [DATA,File,ChTbl] = ReadWIN('C:\test.win','C:\channels.tbl',{'00A0','00B1'},{[2008 07 18 00 00 00],[2008 07 18 00 10 00]});
%
% *** INPUT Parameters ***
% WINfn    : Full Path to the WIN file (char)
% ChTBLfn  : Full Path to the channel table file (char)
% ChIDcell : Channel ID(s) to be imported from the WIN file (cell array)
%          : Each channel ID must be described in hexadecimal (char).
%          : If ChIDcell is empty or omitted, all the channels data are imported.
%          : Ex) {'00A0','00B1'} or []
% StEdTime : Start and end time to be extracted. Ex) {[StartTime],[EndTime]} (cell array), or {[StartTime],DataLength(sec)} or DataLength(sec) or []
%          : Time format is [yyyy mm dd HH MM SS]. Default is whole time period.
%          : Ex) {[2008 07 18 00 00 00],[2008 07 18 00 10 00]}
%          : Ex) {[2008 07 18 00 00 00],600}    <- 600sec from 2008-07-18 00:00:00
%          : Ex) 600                            <- 600sec from the beginning
%
% *** OUTPUT Parameters ***
% DATA    : Waveform data. Column corresponds to channel ID(s).
% File    : File information (struct array)
% ChTble  : Channel table information (struct array)
%
% *** NOTICE1 ***
% This program finds the contained channel ID(s) in the WIN file from the 
% first second header. So the first second data block must involve your 
% desired channel ID data.
%
% *** UPDATES ***
% 2008/July/18(Fri) : Original by Koji MIYAKAWA
%

% ---------------------------
% Parameter Initialization
%
FORMAT = {'bit4','bit8','bit16','bit24','bit32'};

% ---------------------------
% Read WIN Header
%
File.name   = WINfn;
[fid, message] = fopen(File.name,'r','b');
if fid == -1, error(message); end
BLOCK   = 0;
SECROOP = 1;
while 1
    BLOCK = BLOCK + fread(fid,1,'ubit32');                      % Block Size / 4Byte
    if feof(fid), break, end
    HEADtime = (reshape(fread(fid,12,'ubit4'),2,6)'*[10;1])';  	% Channel Header / 4Byte
%   fprintf('SECROOP:%4d, BLOCK:%6d, StartTime:%02d/%02d/%02d %02d:%02d:%02d\n',SECROOP,BLOCK,HEADtime);
    if SECROOP == 1                                             % Read First Second Block
        File.StNum = datenum([HEADtime(1)+CheckYY00(HEADtime(1)),HEADtime(2:end)]);
        [File.StVec,File.StDate,File.StTime] = Num2VecDateTime(File.StNum);
        chn         = 1;
        while ftell(fid) < BLOCK
            HEADch   = fread(fid,1,'ubit16');                   % Channel ID / 2Byte
            HEADsize = fread(fid,1,'ubit4');                    % Sample Size / 0.5Byte
            HEADfrq  = fread(fid,1,'ubit12');                   % Sampling Freqency / 1.5Byte
            DATAtop  = fread(fid,1,'bit32');                    % First value of Sample Data
            DATAdiff = fread(fid,HEADfrq-1,FORMAT{HEADsize+1}); % Differencial values of 1s Sample Data
            if HEADsize == 0 && rem(HEADfrq,2) == 0             % 4 epmty bits are added if sample size is 0.5Byte and sampling rate is even value.
                fread(fid,1,'bit4');
            end
%           fprintf('chn:%d, ChNum:%4X, SampleSize:%-5s, SamplingRate:%4d, ftell(fid):%5d\n',chn,HEADch,char(FORMAT(HEADsize+1)),HEADfrq,ftell(fid));
            ChIDdecTmp(chn) = HEADch;
            SmpFrqTmp(chn)  = HEADfrq;
            chn = chn + 1;
        end
    end
    SECROOP = SECROOP + 1;
    fseek(fid,BLOCK,-1);
end

% ---------------------------
% Start/End Time Setup
%
File.RecSec = round((datenum([HEADtime(1)+CheckYY00(HEADtime(1)),HEADtime(2:end)])-File.StNum)*86400+1);
File.EdNum  = File.StNum+File.RecSec/86400;
[File.EdVec,File.EdDate,File.EdTime] = Num2VecDateTime(File.EdNum);
fprintf('Start time of this WIN file           : %s %s\n',File.StDate,File.StTime);
fprintf('End   time of this WIN file           : %s %s\n',File.EdDate,File.EdTime);
if SECROOP-1 ~= File.RecSec
        disp(['WARNING: This WIN file has packet lost: PacketTotal=',num2str(SECROOP-1),'(sec) < Duration=',num2str(File.RecSec),'(sec)']);
end
if exist('StEdTime','var') && ~isempty(StEdTime)        % Check StEdTime
    if length(StEdTime) == 1                            % StEdTime = DataLength (sec)
        if StEdTime > File.RecSec
            error(['Data length you specified is larger than that of this WIN file : ',num2str(StEdTime),' > ',num2str(File.RecSec),'(sec)']);
        else
            File.RecSec = StEdTime;
            File.EdNum  = File.StNum+File.RecSec/86400;
            [File.EdVec,File.EdDate,File.EdTime] = Num2VecDateTime(File.EdNum);
        end
    elseif length(StEdTime{2}) == 1                     % StEdTime = {[StTime],DataLength(sec)}
        StNum     = datenum(StEdTime{1});
        MaxRecSec = round((File.EdNum-StNum)*86400);    % Maximum Record Length estimated from the start time you specified
        if StNum < File.StNum || StNum > File.EdNum || StEdTime{2} > MaxRecSec
            error(['The variable ''StEdTime'' you specified exceeds the time range of this WIN file : StTime = ',datestr(StEdTime{1},'yyyy/mm/dd('),datestr(StEdTime{1},'ddd) HH:MM:SS'),', Length = ',num2str(StEdTime{2}),'(sec)']);
        else
            File.StNum  = StNum;
            [File.StVec,File.StDate,File.StTime] = Num2VecDateTime(File.StNum);
            File.RecSec = StEdTime{2};
            File.EdNum  = StNum+StEdTime{2}/86400;
            [File.EdVec,File.EdDate,File.EdTime] = Num2VecDateTime(File.EdNum);
        end
    else                                                % StEdTime = {[StTime],[EdTime]}
        StNum  = datenum(StEdTime{1});
        EdNum  = datenum(StEdTime{2});
        if StNum < File.StNum || StNum > File.EdNum || EdNum > File.EdNum
            error(['The variable ''StEdTime'' you specified exceeds the time range of this WIN file : StTime = ',datestr(StEdTime{1},'yyyy/mm/dd('),datestr(StEdTime{1},'ddd) HH:MM:SS'),', EdTime = ',datestr(StEdTime{2},'yyyy/mm/dd('),datestr(StEdTime{2},'ddd) HH:MM:SS')]);
        else
            File.StNum  = StNum;
            [File.StVec,File.StDate,File.StTime] = Num2VecDateTime(File.StNum);
            File.EdNum  = EdNum;
            [File.EdVec,File.EdDate,File.EdTime] = Num2VecDateTime(File.EdNum);
            File.RecSec = round((EdNum-StNum)*86400);
        end
    end
    fprintf('Start time to be extracted            : %s %s\n',File.StDate,File.StTime);
    fprintf('End   time to be extracted            : %s %s\n',File.EdDate,File.EdTime);
end

% ---------------------------
% ChID Setup
%
[ChIDdecTmp,SmpFrqTmp,NumChs,StChs,ColChs,RowChs] = SortChIDdecSmpFrq(ChIDdecTmp,SmpFrqTmp);
fprintf('ChID(s) in this WIN file (%02dch(s))    : %s\n',length(ChIDdecTmp),sprintf('%04X(%GHz) ',[ChIDdecTmp;SmpFrqTmp]));
if exist('ChIDcell','var') && ~isempty(ChIDcell)
    for ii = 1:length(ChIDcell)
        jj = find(ChIDdecTmp == hex2dec(ChIDcell{ii}),1);
        if isempty(jj)
            error(['Your specified ChID (',ChIDcell{ii},') does not exist in the WIN file.']);
        else
            ChIDdec(ii) = ChIDdecTmp(jj);
            SmpFrq(ii)  = SmpFrqTmp(jj);
        end
    end
    [ChIDdec,SmpFrq,NumChs,StChs,ColChs,RowChs] = SortChIDdecSmpFrq(ChIDdec,SmpFrq);
    fprintf('ChID(s) to be extracted  (%02dch(s))    : %s\n',length(ChIDdec),sprintf('%04X(%GHz) ',[ChIDdec;SmpFrq]));
else
    ChIDdec = ChIDdecTmp;
    SmpFrq  = SmpFrqTmp;
end
File.SmpFrq = SmpFrq(StChs);
File.NumChs = NumChs;
File.StChs  = StChs;
File.ColChs = ColChs;
File.RowChs = RowChs;

% ---------------------------
% Read Channel Table File
%
if exist('ChTBLfn','var') && ~isempty(ChTBLfn)
    File.ChTbl = ChTBLfn;
    [fid1, message] = fopen(File.ChTbl,'r','n');
    if fid1 == -1, error(message); end
    while 1
        l = fgetl(fid1);
        if ~ischar(l), break, end
        L  = strread(l,'%s',18,'commentstyle','shell');
        if isempty(L), continue, end
        ii = find(ChIDdec == hex2dec(L{1}),1);
        if isempty(ii), continue, end
        if length(L) < 18
            L(length(L)+1:18,1) = cell(18-length(L),1);     % Zero Padding
        end
        ChTbl.ChID{ii}   = sprintf('%04X',hex2dec(L{1}));
        ChTbl.StnNm{ii}  = L{4};
        ChTbl.Comp{ii}   = L{5};
        ChTbl.ADbit(ii)  = str2double(L{7});
        ChTbl.Sensi(ii)  = str2double(L{8});
        ChTbl.Unit{ii}   = L{9};
        ChTbl.NatFrq(ii) = str2double(L{10});
        ChTbl.Damp(ii)   = str2double(L{11});
        ChTbl.dB(ii)     = str2double(L{12});
        ChTbl.ADres(ii)  = str2double(L{13});
        ChTbl.Lat(ii)    = str2double(L{14});
        ChTbl.Lon(ii)    = str2double(L{15});
        ChTbl.Alt(ii)    = str2double(L{16});
        ChTbl.Pcorr(ii)  = str2double(L{17});
        ChTbl.Scorr(ii)  = str2double(L{18});
    end
    fclose(fid1);
    if ~exist('ChTbl','var') || any(cellfun('isempty',ChTbl.ChID)) || ~all(hex2dec(ChTbl.ChID)'==ChIDdec)
        error('Specified ChID(s) are not found in the ChTbl file.');
    end
else
    File.ChTbl = '';
    for ii = 1:length(ChIDdec)
        ChTbl.ChID{ii}   = sprintf('%04X',ChIDdec(ii));
        ChTbl.StnNm{ii}  = '';
        ChTbl.Comp{ii}   = '';
        ChTbl.ADbit(ii)  = 0;
        ChTbl.Sensi(ii)  = 1;
        ChTbl.Unit{ii}   = 'DigitalCount';
        ChTbl.NatFrq(ii) = 0;
        ChTbl.Damp(ii)   = 0;
        ChTbl.dB(ii)     = 0;
        ChTbl.ADres(ii)  = 1;
        ChTbl.Lat(ii)    = 0;
        ChTbl.Lon(ii)    = 0;
        ChTbl.Alt(ii)    = 0;
        ChTbl.Pcorr(ii)  = 0;
        ChTbl.Scorr(ii)  = 0;
    end
end

% ---------------------------
% Read WIN Data
%
DATA = cell(length(File.NumChs),1);
for ii = 1:length(DATA)
    DATA{ii} = zeros(File.RecSec*File.SmpFrq(ii),File.NumChs(ii))+0.1; % x-axis = (0:size(DATA,1)-1)'/File.SmpFrq;
end
frewind(fid);
BLOCK   = 0;
while 1
    BLOCK = BLOCK + fread(fid,1,'ubit32');              % Block Size / 4Byte
    if feof(fid), break, end
    HEADtime = (reshape(fread(fid,12,'ubit4'),2,6)'*[10;1])';	% Channel Header / 4Byte
    CurNum   = datenum([HEADtime(1)+CheckYY00(HEADtime(1)),HEADtime(2:end)]);
    CulPos   = round((CurNum-File.StNum)*86400*File.SmpFrq+1);  % Culumn Position of the start point of this second roop data
    if CulPos(1) < 1 || CulPos(1) > File.RecSec*File.SmpFrq(1) % Skip this second block or not
        fseek(fid,BLOCK,-1);
        continue;
    end
    while ftell(fid) < BLOCK
        HEADch = fread(fid,1,'ubit16');                 % Channel ID / 2Byte
        chn    = find(ChIDdec == HEADch,1);
        if isempty(chn)
            HEADsize = fread(fid,1,'ubit4');            % Sample Size / 0.5Byte
            HEADfrq  = fread(fid,1,'ubit12');           % Sampling Freqency / 1.5Byte
            fread(fid,1,'bit32');                       % First Data
            fread(fid,HEADfrq-1,FORMAT{HEADsize+1});    % Differencial Data
            if HEADsize == 0 && rem(HEADfrq,2) == 0     % 4 epmty bits are added if sample size is 0.5Byte and sampling rate is even value.
                fread(fid,1,'bit4');
            end
%           fprintf('chn:%d, ChNum:%4X, SampleSize:%-5s, SamplingRate:%4d, ftell(fid):%5d\n',chn,HEADch,char(FORMAT(HEADsize+1)),HEADfrq,ftell(fid));
        else
            HEADsize = fread(fid,1,'ubit4');            % Sample Size / 0.5Byte
            HEADfrq  = fread(fid,1,'ubit12');           % Sampling Freqency / 1.5Byte
            DATA{File.ColChs(chn)}(CulPos(File.ColChs(chn)):CulPos(File.ColChs(chn))+HEADfrq-1,File.RowChs(chn)) = cumsum([fread(fid,1,'bit32');fread(fid,HEADfrq-1,FORMAT{HEADsize+1})]); % Cumulative Sum
            if HEADsize == 0 && rem(HEADfrq,2) == 0     % 4 epmty bits are added if sample size is 0.5Byte and sampling rate is even value.
                fread(fid,1,'bit4');
            end
%           fprintf('chn:%d, ChNum:%4X, SampleSize:%-5s, SamplingRate:%4d, ftell(fid):%5d\n',chn,HEADch,char(FORMAT(HEADsize+1)),HEADfrq,ftell(fid));
        end
    end
end
fclose(fid);

% ---------------------------
% For MeSO-net Temperature and Barometric Pressure Data
%
for ii = 1:length(DATA)
    if File.SmpFrq(ii) == 1 && (strcmpi(ChTbl.Comp(File.StChs(ii)),'TE') || strcmpi(ChTbl.Comp(File.StChs(ii)),'AP'))
        fprintf('1Hz data are decimated to 1 minute    : %s\n',sprintf('%s ',ChTbl.ChID{find(File.ColChs==ii,1)}));
        File.SmpFrq(ii) = 1/60;             % Set sampling frequency to 1 minute
        DATA{ii} = DATA{ii}(1:60:end,:);
    end
end

% ---------------------------
% Replace lost packet data(=0) with the one point previous data
%
for ii = 1:length(DATA)
    File.NoDATA{ii} = find(any(DATA{ii}(1:ceil(File.SmpFrq(ii)):end,:)==0.1,2)')-1;
    if ~isempty(File.NoDATA{ii})
        fprintf('WARNING: Packet lost(sec) on %4GHz   : ',File.SmpFrq(ii)); fprintf('%d ',File.NoDATA{ii}); fprintf('\n');
    end
    [I,J] = find(DATA{ii}==0.1);
    for jj = 1:length(I)
        if I(jj) ~= 1   % First data point can not be replaced
            DATA{ii}(I(jj),J(jj)) = DATA{ii}(I(jj)-1,J(jj));
        end
    end
end

% ---------------------------
% Data Unit Conversion
%
if ~isempty(File.ChTbl)
    for ii = 1:sum(File.NumChs)
            DATA{File.ColChs(ii)}(:,File.RowChs(ii)) = DATA{File.ColChs(ii)}(:,File.RowChs(ii))/10^(ChTbl.dB(ii)/20)*ChTbl.ADres(ii)/ChTbl.Sensi(ii);
    end
end

% ---------------------------
% Add Carriage Return
%
fprintf('\n');

% ===========================
function [ChIDdec,SmpFrq,NumChs,StChs,ColChs,RowChs]=SortChIDdecSmpFrq(ChIDdec,SmpFrq)

[SmpFrq,I] = sort(SmpFrq,'descend');
ChIDdec    = ChIDdec(I);
ColChs     = zeros(size(ChIDdec));
RowChs     = zeros(size(ChIDdec));
ROOP       = 1;
ii         = 1;
while 1
    INDEX = find(SmpFrq==SmpFrq(ii));
    ChIDdec(INDEX)  = sort(ChIDdec(INDEX));
    NumChs(ROOP) = length(INDEX);
    StChs(ROOP)  = INDEX(1);
    ColChs(INDEX) = ROOP;
    RowChs(INDEX) = 1:length(INDEX);
    ii   = INDEX(end)+1;
    if ii > length(ChIDdec)
        break
    end
    ROOP = ROOP+1;
end

% ===========================
function [Vec,Date,Time]=Num2VecDateTime(Num)
% Convert a serial date number to vector, date string, and time string
%
Vec  = datevec(Num);
Date = [datestr(Num,'yyyy/mm/dd('),datestr(Num,'ddd)')];
Time = datestr(Num,'HH:MM:SS');

% ===========================
function YY00 = CheckYY00(yy)
% Check YY00 is 1900 or 2000
%
if yy < 50
	YY00 = 2000;
else
    YY00 = 1900;
end
